/*
 * This file (snpost.c) is part of the sn package.
 * Distribution of sn is covered by the GPL. See file COPYING.
 * sn is copyright (c) 1998 Harold Tay.
 */
/*
 * Post articles to a news server.  Network is read on fd 6, write
 * on fd 7.
 */

#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <utime.h>
#include <string.h>
#include <time.h>
#include <dirent.h>
#include <errno.h>
#include <stdarg.h>
#include "config.h"
#include "parameters.h"
#include "args.h"
#include "head.h"
#include <b.h>
#include <out.h>
#include <readln.h>
#include <format.h>
#include <cmdopen.h>
#include <opt.h>

int debug = 0;

static const char rcsid[] = "$Id: snpost.c,v 1.22 1999/03/27 05:28:42 harold Exp $";

static void myunlink(char * fn) { if( unlink(fn) )LOG("unlink(%s):%m", fn); }

struct readln input;

static int lines, bytes;

static int
switchline(char * line)
{
  int len;

  len = strlen(line);
  lines++;
  bytes += len;
  switch( *line ){
  case 'n': case 'N':
    if( 0 == strncasecmp(line, "NNTP-Posting-Host:", 18) )return(1);
    if( 0 == strncasecmp(line, "Newsgroups:", 11) )break;
    return(0);
  case 'f': case 'F':
    if( 0 == strncasecmp(line, "From:", 5) )break;
    return(0);
  case 's': case 'S':
    if( 0 == strncasecmp(line, "Subject:", 8) )break;
  default:
    return(0);
  }
  log("Article %S", len-2, line);
  return(0);
}

static void nomem(void) { fail(2, "No memory"); }
static void writeerr(void) { fail(2, "network read/write error:%m"); }
static void toolong(void) { fail(3, "server reply too long:%S...", 20, args_inbuf); }

static int
post(int fd)
{
  struct b head;
  int er;
  char * buf;
  int len;
  struct readln r;
  int n;

  if( -1 == readln_ready(fd, 0, &r) )nomem();

  bytes = lines = head.used = head.size = 0;
  head.buf = 0;

  switch( read_unfold_head(&head, &r, switchline, "\r\n") ){
  case -2: case 1: case 2:
    log("post:article in bad format"); readln_done(&r); return(1);
  case -1:
    if( ENOMEM == errno )nomem();
    writeerr();
  default:
    if( head.used )break;
    log("post:empty article"); readln_done(&r); return(0);
  }

  if( -1 == args_write(7, "POST\r\n") )writeerr();
  if( -1 == (er = args_read(&input)) )writeerr();
  if( 0 == er )toolong();
  if( atoi(args[0]) != 340 ){
    args_report("post:"); return(1); }

  if( -1 == b_appendl(&head, "\r\n", 2) )nomem();
  if( -1 == write(7, head.buf, head.used) )writeerr();
  free(head.buf);

  lines++;
  while( (len = readln(&r, &buf, '\n')) > 0 ){
    lines++;
    bytes += len;
    if( -1 == write(7, buf, len) )writeerr();
    if( 3 == len && '.' == *buf && '\r' == buf[1] )
      break;
  }

  if( len != 3 || '.' != *buf ){
    LOG1("post:Article does not end with dot");
    if( -1 == write(7, "\r\n.\r\n", 5) )writeerr();
  }
  readln_done(&r);

  if( -1 == (er = args_read(&input)) )writeerr();
  if( 0 == er )toolong();
  if( (n = atoi(args[0]) - 240) ){
    args_report("post:posting failed:");
    return(1);
  }
  LOG1("post: ok, %d lines, %d bytes", lines, bytes);
  return(0);
}

static void
post_error(int fd, char * fmt, ...)
{
  int pid;
  int p;
  int i;
  char * line;
  static char * c[] = {"mail", "-s", "Posting failed", 0, 0 };
  struct readln r;
  va_list ap;

  va_start(ap, fmt);
  if( ! c[3] )
    if( ! (c[3] = getenv("LOGNAME")) || ! *c[3] )
      c[3] = NEWSMASTER;

  putenv("LOGNAME=sn");
  putenv("MAILNAME=sn");
  putenv("QMAILUSER=sn");
  if( (pid = cmdopen(c, 0, &p)) > -1 ){
    lseek(fd, 0, SEEK_SET);
    if( 0 == readln_ready(fd, 0, &r) ){
      writes(p, "This is message generated by the snpost program, don't reply.\n\n");
      writefv(p, fmt, ap);
      writes(p, "The article follows:\n\n");
      /* the article already dot-stuffed by orig poster */
      while( (i = readln(&r, &line, '\n')) > 0 )
        writef(p, "%S\n", i-2, line);
      readln_done(&r);
    }else
      log("post_error:readln_ready:%m");
    close(p);
    if( (i = cmdwait(pid)) )LOG("post_error:mail command died with %d", i);
  }else
    log("post_error:cmdopen:%m?");
  close(fd);
  va_end(ap);
}

static char *
gettime(void)
{
  struct stat st;
  static char buf[16];
  struct tm * tmp;
  char * p;

  if( -1 == stat(".last-list", &st) )
    return("000000 000000");

  tmp = localtime(&st.st_mtime);
  if( tmp->tm_year > 100 )tmp->tm_year -= 100;
  tmp->tm_mon += 1;
  p = buf;
  *p++ = '0' + tmp->tm_year/10; *p++ = '0' + tmp->tm_year%10;
  *p++ = '0' + tmp->tm_mon/10;  *p++ = '0' + tmp->tm_mon%10;
  *p++ = '0' + tmp->tm_mday/10; *p++ = '0' + tmp->tm_mday%10;
  *p++ = ' ';
  *p++ = '0' + tmp->tm_hour/10; *p++ = '0' + tmp->tm_hour%10;
  *p++ = '0' + tmp->tm_min/10;  *p++ = '0' + tmp->tm_min%10;
  *p++ = '0' + tmp->tm_sec/10;  *p++ = '0' + tmp->tm_sec%10;
  *p = '\0';
  return(buf);
}

static int
docommand(char * send, int expect, char * exec)
{
  int pid;
  int p;
  int er;

  if( (pid = cmdopensh(exec, 0, &p)) <= 0 ){
    LOG("docommand:fork/exec sh -c \"%s\": %m"); return(-1); }
  args_write(7, "%s\r\n", send);
  if( -1 == (er = args_read(&input)) )
    LOG("docommand:args_read:%m");
  if( 0 == er )toolong();
  else if( expect != atoi(args[0]) ){
    args_report("docommand:");
    args_flushtodot(&input);
  }else{
    int len;
    char * buf;
    writef(p, "->%s\n<-%s\n", args_outbuf, args_inbuf);
    while( (len = readln(&input, &buf, '\n')) > 0 ){
      write(p, buf, len - 2);
      write(p, "\n", 1);
      if( 3 == len && '.' == *buf && '\r' == buf[1] )
        break;
    }
    if( len <= 0 ){
      writes(p, "\n\n[Server failed]\n");
      log("Failed \"%s\"", send);
      er = -1;
    }else{
      writes(p, "\n\n[Succeeded]\n");
      log("Did \"%s\"", send);
    }
  }
  close(p);
  if( (er = cmdwait(pid)) )LOG("command \"%s\" died with %d", exec, er);
  return(er);
}

static char *
opencheckmoderead(char * fn, int * fdp)
{
  struct stat st;
  static char cmdbuf[1024];
  int fd;

  if( -1 == (fd = open(fn, O_RDONLY)) ){
    if( ENOENT != errno )LOG("opencheckmoderead:open(%s):%m", fn);
    return(NULL);
  }
  if( -1 == flock(fd, LOCK_EX|LOCK_NB) )
    ;
  else if( -1 == fstat(*fdp = fd, &st) )
    LOG("opencheckmoderead:fstat(%s):%m", fn);
  else if( st.st_uid != 0 && st.st_uid != snuid)
    LOG("opencheckmoderead:%s has wrong owner", fn);
  else if( st.st_mode & 022 )
    LOG("opencheckmoderead:%s is writable by non-owner", fn);
  else if( st.st_size > sizeof(cmdbuf)-1 )
    LOG("opencheckmoderead:%s too big", fn);
  else if( -1 == read(fd, cmdbuf, st.st_size) )
    LOG("opencheckmoderead:read(%s):%m", fn);
  else{
    cmdbuf[st.st_size] = '\0';
    return(cmdbuf);
  }
  close(fd); return(NULL);
}

void
commands(void)
{
  char * buf;
  char c[128] = "NEWGROUPS ";
  int fd;
  int a, b;

  a = b = 1;
  if( (buf = opencheckmoderead("request-newgroups", &fd)) ){
    strcat(c, gettime());
    if( 0 == (a = docommand(c, 231, buf)) ){
      myunlink("request-newgroups");
      if( debug > 2 )log("request-newgroups ok");
    }
    close(fd);
  }

  if( (buf = opencheckmoderead("request-list", &fd)) ){
    if( 0 == (b = docommand("LIST NEWSGROUPS", 215, buf)) ){
      myunlink("request-list");
      if( debug > 2 )log("request-list ok");
    }
    close(fd);
  }

  if( 0 == a || 0 == b )utime(".last-list", 0);
}

static void usage(void) {
  fail(1, "Usage:%s [-t timeout] newsgroup", progname); }

char servnam[MAXPATHLEN + 1];

int
main(int argc, char ** argv)
{
  time_t now;
  char * newsgroup = NULL;
  char * cp;
  DIR * dir;
  struct dirent * dp;
  int i;
  int tmo = 120;
  int posted = 0;
  int logpid = 0;

  progname = ((cp = strrchr(argv[0], '/'))?cp + 1:argv[0]);

  parameters(1);

  while( (i = opt_get(argc, argv, "t")) > -1 )
    switch( i ){
    case 'd': debug++; break;
    case 'V': version(); exit(0);
    case 't': if( ! opt_arg )usage(); tmo = atoi(opt_arg); break;
    case 'P': logpid = 1; break;
    default: usage();
    }
  if( opt_ind >= argc )usage();
  newsgroup = argv[opt_ind++];
  if( opt_ind < argc )usage();

  if( chdir(snroot) || chdir(newsgroup) )
    fail((errno==ENOENT)?1:2, "Can't chdir(%s/%s):%m", snroot, newsgroup);
  if( -1 == (i = readlink(".outgoing", servnam, sizeof(servnam)-1)) )
    fail((errno==ENOENT)?1:2, "readlink(.outgoing):%m");
  servnam[i] = '\0';
  if( chdir(servnam) )
    fail(2, "chdir(%s/%s/%s):%m", snroot, newsgroup, servnam);

  if( -1 == readln_ready(6, tmo, &input) )FAIL(2, "readln_ready:%m");

  if( (cp = malloc(i = strlen(progname) + i + 32)) ){
    char * s = strrchr(servnam, '/');
    if( s )s++; else s = servnam;
    if( ! logpid )formats(cp, i-1, "%s:%s", progname, s);
    else formats(cp, i-1, "%s:%s[%u]", progname, s, getpid());
    progname = cp;
  }

  if( (dir = opendir(".")) ){
    struct stat st;
    if( -1 == fstat(dirfd(dir), &st) )fail(2, "fstat(.):%m");
    if( st.st_uid != snuid || st.st_gid != sngid )
      fail(1, "wrong ownership on directory");
    if( st.st_mode&02 )
      fail(1, "directory is world writeable");
  }else
    FAIL(2, "opendir(.):%m");

  commands();

  if( ! (dp = readdir(dir)) )FAIL(2, "readdir:%m");
  time(&now);
  do{
    struct stat st;
    int fd;

    if( '$' != *dp->d_name )continue;
    if( -1 == (fd = open(dp->d_name, O_RDWR)) ){
      LOG("open(%s)/rw:%m", dp->d_name); continue; }
    if( -1 == fstat(fd, &st) )LOG("fstat(%s):%m", dp->d_name);
    else if( ! S_ISREG(st.st_mode) );
    else if( -1 == flock(fd, LOCK_EX|LOCK_NB) ){
      if( EWOULDBLOCK != errno )LOG("flock(%s):%m", dp->d_name); 
    }else if( now - st.st_mtime > MAX_POST_AGE ){
      log("Deleting old post %s", dp->d_name);
      post_error(fd, "This message is too old.\n\n");
      myunlink(dp->d_name);
    }else{
      if( 0 == (i = post(fd)) )posted++;
      else post_error(fd, "An error occurred while posting this"
        " article.\nI said: \"%s\"\nThe server replied: \"%s\"\n\n",
        args_outbuf, args_inbuf);
      myunlink(dp->d_name);
    }
    close(fd);
  }while( (dp = readdir(dir)) );

  closedir(dir);
  if( posted > 0 )log("posted %d articles", posted);
  exit(0);
}
